/*
 * The Rubik's cube.
 *
 * Sed - april 1999 / december 2003.
 *
 * This program is in the public domain.
 *--------------------
 * The main file.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "messages.h"
#include "screen.h"
#include "cube.h"
#include "event.h"
#include "device.h"
#include "version.h"

device d;

char *the_screen;

double EDGE;
char *FONT;
int HELP_X;
int HELP_Y;
int FORMAT = 0;

int SCREEN_X;
int SCREEN_Y;
double CUBE_SIZE;
double PERSPECTIVE;

/* J.-P. Demailly's coloring way... */
char *palette[11] = {
   "grey50",          /* face X (empty) */
   "green",           /* face A */
   "red",             /* face B */
   "royalblue1",      /* face C */
   "yellow",          /* face D */
   "orange",          /* face E */
   "ivory",           /* face F */
   "grey24",          /* internal faces */
   "red4",            /* cross & letters color */     
   "white",           /* foreground color */     
   "black",           /* background color */
};

char *formats[] = { "position", "numeric", "generators"};

void usage()
{
  int i;
   
  fprintf(stderr, "\n");
  for (i=0; i<help_numlines; i++)
      fprintf(stderr, "%s\n", help_index[i]);
}

int main(int nb, char **argv)
{
  CUBE cube;
  SCREEN screen;
   
  int i, j;
  int rand=1;
  char *filename = NULL;
  char *ptr;
   
  FONT = "7x13";
  EDGE = 0.475;

  SCREEN_X = 320;
  SCREEN_Y = 360;
  PERSPECTIVE = 0.3;  /* 0.2 changed to 0.3, I like it better - Sed */
   
  ptr = help_msg;
  help_numlines = 1;
  i = 0;
  help_index = (char **) malloc(sizeof(char *));
  help_index[0] = ptr;
   
  while(*ptr) {
    if (*ptr=='\n') {
      *ptr = '\0';
      help_index = (char **) realloc(help_index,
			             (help_numlines+1)*sizeof(char *));       
      help_index[help_numlines] = ptr+1;
      ++help_numlines;
      if (i>help_maxlen) help_maxlen = i;
      i = 0;
    }
    ++ptr;
    ++i;
  }
	
  the_screen=&screen.buffer[0];
  d.buffer=screen.buffer;
  screen.buffer = NULL;
  screen.im = 0;
   
  for (i=1; i<nb; i++) {
    if (!strcasecmp(argv[i], "-h") || !strcasecmp(argv[i], "--h") || !strcasecmp(argv[i], "-help") ||
	!strcasecmp(argv[i], "--help")) {
      usage();
      return 0;
    }
    if (!strcasecmp(argv[i], "-i") || !strcasecmp(argv[i], "--i") || !strcasecmp(argv[i], "-init") || !strcasecmp(argv[i], "--init")) {
      rand = 0;
      continue;
    }
    if (!strcasecmp(argv[i], "-v") || !strcasecmp(argv[i], "--v") || !strcasecmp(argv[i], "-version") ||
	!strcasecmp(argv[i], "--version")) {
      fprintf(stderr, "%s\n", VERSION);
      return 0;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-e") ||
		   !strcasecmp(argv[i], "--e") ||
		   !strcasecmp(argv[i], "-edge") ||
		   !strcasecmp(argv[i], "--edge"))) {
      EDGE = 0.5 * atof(argv[++i]);
      if (EDGE<0.1) EDGE = 0.1;
      if (EDGE>0.5) EDGE = 0.5;
      continue;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-l") ||
		   !strcasecmp(argv[i], "--l") ||
		   !strcasecmp(argv[i], "-load") ||
		   !strcasecmp(argv[i], "--load"))) {
      filename = argv[++i];
      continue;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-fn") ||
		   !strcasecmp(argv[i], "--fn") ||
		   !strcasecmp(argv[i], "-font") ||
		   !strcasecmp(argv[i], "--font"))) {
      FONT = argv[++i];
      continue;
    }     
    if (i<nb-1 && (!strcasecmp(argv[i], "-fm") ||     
		   !strcasecmp(argv[i], "--fm") ||
		   !strcasecmp(argv[i], "-format") ||
		   !strcasecmp(argv[i], "--format"))) {
      FORMAT = atoi(argv[++i]);
      if (FORMAT<0 || FORMAT>2) FORMAT = 2;
      continue;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-width") || 
		   !strcasecmp(argv[i], "--width"))) {
      SCREEN_X = atoi(argv[++i]);
      if (SCREEN_X<80) SCREEN_X = 80;
      if (SCREEN_X>1600) SCREEN_X = 1600;
      continue;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-height") ||
		   !strcasecmp(argv[i], "--height"))) {
      SCREEN_Y = atoi(argv[++i]);
      if (SCREEN_Y<60) SCREEN_Y = 60;
      if (SCREEN_Y>1200) SCREEN_Y = 1200;
      continue;
    }
    if (i<nb-2 && (!strcasecmp(argv[i], "-color") ||
		   !strcasecmp(argv[i], "--color"))) {
      ++i;
      if (argv[i][0]=='x' || argv[i][0]=='X') j = 0;
        else	 	 
      if (argv[i][0]=='i' || argv[i][0]=='I') j = 7;
        else	 
      if (argv[i][0]=='0') j = 10;
        else
      if (argv[i][0]=='1') j = 9;
        else
      if (argv[i][0]=='2') j = 8;
        else
      j = toupper(argv[i][0])-'A' + 1;
      if (j>=0 && j<=10) palette[j] = argv[++i];
      continue;
    }
    fprintf(stderr, "unknown option %s\n", argv[i]);
    usage();
    return 1;
  }
  set_cube_size();
   
  if (init_screen(&screen)==-1) {
    fprintf(stderr, "Error with initing the screen.\n");
    return -1;
  }

  reset_cube(&cube);

  if (filename)
    init_cube(&cube, filename, 0);
  else
    init_cube(&cube, NULL, rand);

  realize_colors(&screen);
  screen.pause = 1;   
  grab_pointer(&screen, 0);
  display_cube(&cube, &screen);

  while(1) {   
    int d=wait_event(&screen, &cube);
     
    if (!(d&SCREEN_EVENT)) {
      if (cube.anim)
	display_cube(&cube, &screen);
    } else {
      if (screen_event(&screen, &cube)==-1)
	break;
    }
  }

  close_screen(&screen);

  return 0;
}
